# Author: Nicolas VERDIER
# This file is part of memorpy.
#
# memorpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# memorpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with memorpy.  If not, see <http://www.gnu.org/licenses/>.

from . import utils


class AddressException(Exception):
    pass


class Address:
    """this class is used to have better representation of memory addresses"""

    def __init__(self, value, process, default_type="uint"):
        self.value = int(value)
        self.process = process
        self.default_type = default_type
        self.symbolic_name = None

    def read(self, type=None, maxlen=None, errors="raise"):
        if maxlen is None:
            try:
                int(type)
                maxlen = int(type)
                type = None
            except:
                pass

        if not type:
            type = self.default_type
        if not maxlen:
            return self.process.read(self.value, type=type, errors=errors)
        else:
            return self.process.read(
                self.value, type=type, maxlen=maxlen, errors=errors
            )

    def write(self, data, type=None):
        if not type:
            type = self.default_type
        return self.process.write(self.value, data, type=type)

    def symbol(self):
        return self.process.get_symbolic_name(self.value)

    def get_instruction(self):
        return self.process.get_instruction(self.value)

    def dump(self, ftype="bytes", size=512, before=32):
        buf = self.process.read_bytes(self.value - before, size)
        print(utils.hex_dump(buf, self.value - before, ftype=ftype))

    def __nonzero__(self):
        return self.value is not None and self.value != 0

    def __add__(self, other):
        return Address(self.value + int(other), self.process, self.default_type)

    def __sub__(self, other):
        return Address(self.value - int(other), self.process, self.default_type)

    def __repr__(self):
        if not self.symbolic_name:
            self.symbolic_name = self.symbol()
        return str("<Addr: %s" % self.symbolic_name + ">")

    def __str__(self):
        if not self.symbolic_name:
            self.symbolic_name = self.symbol()
        return str(
            "<Addr: %s" % self.symbolic_name
            + ' : "%s" (%s)>'
            % (str(self.read()).encode("unicode_escape"), self.default_type)
        )

    def __int__(self):
        return int(self.value)

    def __hex__(self):
        return hex(self.value)

    def __get__(self, instance, owner):
        return self.value

    def __set__(self, instance, value):
        self.value = int(value)

    def __lt__(self, other):
        return self.value < int(other)

    def __le__(self, other):
        return self.value <= int(other)

    def __eq__(self, other):
        return self.value == int(other)

    def __ne__(self, other):
        return self.value != int(other)

    def __gt__(self, other):
        return self.value > int(other)

    def __ge__(self, other):
        return self.value >= int(other)



#!/usr/bin/env python
# -*- coding: UTF8 -*-

import struct

from .utils import type_unpack

""" Base class for process not linked to any platform """


class ProcessException(Exception):
    pass


class BaseProcess:
    def __init__(self, *args, **kwargs):
        """ Create and Open a process object from its pid or from its name """
        self.h_process = None
        self.pid = None
        self.isProcessOpen = False
        self.buffer = None
        self.bufferlen = 0

    def __del__(self):
        self.close()

    def close(self):
        pass

    def iter_region(self, *args, **kwargs):
        raise NotImplementedError

    def write_bytes(self, address, data):
        raise NotImplementedError

    def read_bytes(self, address, bytes=4):
        raise NotImplementedError

    def get_symbolic_name(self, address):
        return "0x%08X" % int(address)

    def read(self, address, type="uint", maxlen=50, errors="raise"):
        if type == "s" or type == "string":
            s = self.read_bytes(int(address), bytes=maxlen)
            news = ""
            for c in s:
                if c == "\x00":
                    return news
                news += c
            if errors == "ignore":
                return news
            raise ProcessException("string > maxlen")
        else:
            if type == "bytes" or type == "b":
                return self.read_bytes(int(address), bytes=maxlen)
            s, l = type_unpack(type)
            return struct.unpack(s, self.read_bytes(int(address), bytes=l))[0]

    def write(self, address, data, type="uint"):
        if type != "bytes":
            s, l = type_unpack(type)
            return self.write_bytes(int(address), struct.pack(s, data))
        else:
            return self.write_bytes(int(address), data)



# Author: Nicolas VERDIER
# This file is part of memorpy.
#
# memorpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# memorpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with memorpy.  If not, see <http://www.gnu.org/licenses/>.

import re
import struct


def re_to_unicode(s):
    newstring = ""
    for c in s:
        newstring += re.escape(c) + "\\x00"

    return newstring


def type_unpack(type):
    """ return the struct and the len of a particular type """
    type = type.lower()
    s = None
    l = None
    if type == "short":
        s = "h"
        l = 2
    elif type == "ushort":
        s = "H"
        l = 2
    elif type == "int":
        s = "i"
        l = 4
    elif type == "uint":
        s = "I"
        l = 4
    elif type == "long":
        s = "l"
        l = 4
    elif type == "ulong":
        s = "L"
        l = 4
    elif type == "float":
        s = "f"
        l = 4
    elif type == "double":
        s = "d"
        l = 8
    else:
        raise TypeError("Unknown type %s" % type)
    return ("<" + s, l)


def hex_dump(data, addr=0, prefix="", ftype="bytes"):
    """
    function originally from pydbg, modified to display other types
    """
    dump = prefix
    slice = ""
    if ftype != "bytes":
        structtype, structlen = type_unpack(ftype)
        for i in range(0, len(data), structlen):
            if addr % 16 == 0:
                dump += " "
                for char in slice:
                    if ord(char) >= 32 and ord(char) <= 126:
                        dump += char
                    else:
                        dump += "."

                dump += "\n%s%08X: " % (prefix, addr)
                slice = ""
            tmpval = "NaN"
            try:
                packedval = data[i : i + structlen]
                tmpval = struct.unpack(structtype, packedval)[0]
            except Exception as e:
                print(e)

            if tmpval == "NaN":
                dump += "{:<15} ".format(tmpval)
            elif ftype == "float":
                dump += "{:<15.4f} ".format(tmpval)
            else:
                dump += "{:<15} ".format(tmpval)
            addr += structlen

    else:
        for byte in data:
            if addr % 16 == 0:
                dump += " "
                for char in slice:
                    if ord(char) >= 32 and ord(char) <= 126:
                        dump += char
                    else:
                        dump += "."

                dump += "\n%s%08X: " % (prefix, addr)
                slice = ""
            dump += "%02X " % ord(byte)
            slice += byte
            addr += 1

    remainder = addr % 16
    if remainder != 0:
        dump += "   " * (16 - remainder) + " "
    for char in slice:
        if ord(char) >= 32 and ord(char) <= 126:
            dump += char
        else:
            dump += "."

    return dump + "\n"



# Author: Nicolas VERDIER
# This file is part of memorpy.
#
# memorpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# memorpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with memorpy.  If not, see <http://www.gnu.org/licenses/>.

from ctypes import (
    Structure,
    c_long,
    c_int,
    c_uint,
    c_char,
    c_void_p,
    c_ubyte,
    c_ushort,
    c_ulong,
    c_ulonglong,
    windll,
    POINTER,
    sizeof,
    c_bool,
    c_size_t,
    c_longlong,
)
from ctypes.wintypes import *

if sizeof(c_void_p) == 8:
    ULONG_PTR = c_ulonglong
else:
    ULONG_PTR = c_ulong


class SECURITY_DESCRIPTOR(Structure):
    _fields_ = [
        ("SID", DWORD),
        ("group", DWORD),
        ("dacl", DWORD),
        ("sacl", DWORD),
        ("test", DWORD),
    ]


PSECURITY_DESCRIPTOR = POINTER(SECURITY_DESCRIPTOR)


class MEMORY_BASIC_INFORMATION(Structure):
    _fields_ = [
        ("BaseAddress", c_void_p),
        ("AllocationBase", c_void_p),
        ("AllocationProtect", DWORD),
        ("RegionSize", c_size_t),
        ("State", DWORD),
        ("Protect", DWORD),
        ("Type", DWORD),
    ]


# https://msdn.microsoft.com/fr-fr/library/windows/desktop/aa366775(v=vs.85).aspx
class MEMORY_BASIC_INFORMATION64(Structure):
    _fields_ = [
        ("BaseAddress", c_ulonglong),
        ("AllocationBase", c_ulonglong),
        ("AllocationProtect", DWORD),
        ("alignement1", DWORD),
        ("RegionSize", c_ulonglong),
        ("State", DWORD),
        ("Protect", DWORD),
        ("Type", DWORD),
        ("alignement2", DWORD),
    ]


class SYSTEM_INFO(Structure):
    _fields_ = [
        ("wProcessorArchitecture", WORD),
        ("wReserved", WORD),
        ("dwPageSize", DWORD),
        ("lpMinimumApplicationAddress", LPVOID),
        ("lpMaximumApplicationAddress", LPVOID),
        ("dwActiveProcessorMask", ULONG_PTR),
        ("dwNumberOfProcessors", DWORD),
        ("dwProcessorType", DWORD),
        ("dwAllocationGranularity", DWORD),
        ("wProcessorLevel", WORD),
        ("wProcessorRevision", WORD),
    ]


class PROCESSENTRY32(Structure):
    _fields_ = [
        ("dwSize", c_uint),
        ("cntUsage", c_uint),
        ("th32ProcessID", c_uint),
        ("th32DefaultHeapID", c_uint),
        ("th32ModuleID", c_uint),
        ("cntThreads", c_uint),
        ("th32ParentProcessID", c_uint),
        ("pcPriClassBase", c_long),
        ("dwFlags", DWORD),
        # ('dwFlags', ULONG_PTR),
        ("szExeFile", c_char * 260),
        ("th32MemoryBase", c_long),
        ("th32AccessKey", c_long),
    ]


class MODULEENTRY32(Structure):
    _fields_ = [
        ("dwSize", c_uint),
        ("th32ModuleID", c_uint),
        ("th32ProcessID", c_uint),
        ("GlblcntUsage", c_uint),
        ("ProccntUsage", c_uint),
        ("modBaseAddr", c_uint),
        ("modBaseSize", c_uint),
        ("hModule", c_uint),
        ("szModule", c_char * 256),
        ("szExePath", c_char * 260),
    ]


class THREADENTRY32(Structure):
    _fields_ = [
        ("dwSize", c_uint),
        ("cntUsage", c_uint),
        ("th32ThreadID", c_uint),
        ("th32OwnerProcessID", c_uint),
        ("tpBasePri", c_uint),
        ("tpDeltaPri", c_uint),
        ("dwFlags", c_uint),
    ]


class TH32CS_CLASS(object):
    INHERIT = 2147483648
    SNAPHEAPLIST = 1
    SNAPMODULE = 8
    SNAPMODULE32 = 16
    SNAPPROCESS = 2
    SNAPTHREAD = 4
    ALL = 2032639


Module32First = windll.kernel32.Module32First
Module32First.argtypes = [c_void_p, POINTER(MODULEENTRY32)]
Module32First.rettype = c_int
Module32Next = windll.kernel32.Module32Next
Module32Next.argtypes = [c_void_p, POINTER(MODULEENTRY32)]
Module32Next.rettype = c_int

Process32First = windll.kernel32.Process32First
Process32First.argtypes = [c_void_p, POINTER(PROCESSENTRY32)]
Process32First.rettype = c_int
Process32Next = windll.kernel32.Process32Next
Process32Next.argtypes = [c_void_p, POINTER(PROCESSENTRY32)]
Process32Next.rettype = c_int

CreateToolhelp32Snapshot = windll.kernel32.CreateToolhelp32Snapshot
CreateToolhelp32Snapshot.reltype = c_long
CreateToolhelp32Snapshot.argtypes = [c_int, c_int]

CloseHandle = windll.kernel32.CloseHandle
CloseHandle.argtypes = [c_void_p]
CloseHandle.rettype = c_int

OpenProcess = windll.kernel32.OpenProcess
OpenProcess.argtypes = [c_void_p, c_int, c_long]
OpenProcess.rettype = c_long
OpenProcessToken = windll.advapi32.OpenProcessToken
OpenProcessToken.argtypes = (HANDLE, DWORD, POINTER(HANDLE))
OpenProcessToken.restype = BOOL

ReadProcessMemory = windll.kernel32.ReadProcessMemory
ReadProcessMemory.argtypes = [HANDLE, LPCVOID, LPVOID, c_size_t, POINTER(c_size_t)]
ReadProcessMemory = windll.kernel32.ReadProcessMemory

WriteProcessMemory = windll.kernel32.WriteProcessMemory
WriteProcessMemory.argtypes = [HANDLE, LPVOID, LPCVOID, c_size_t, POINTER(c_size_t)]
WriteProcessMemory.restype = BOOL

if sizeof(c_void_p) == 8:
    NtWow64ReadVirtualMemory64 = None
else:
    try:
        NtWow64ReadVirtualMemory64 = windll.ntdll.NtWow64ReadVirtualMemory64
        NtWow64ReadVirtualMemory64.argtypes = [
            HANDLE,
            c_longlong,
            LPVOID,
            c_ulonglong,
            POINTER(c_ulong),
        ]  # NTSTATUS (__stdcall *NtWow64ReadVirtualMemory64)(HANDLE ProcessHandle, PVOID64 BaseAddress, PVOID Buffer, ULONGLONG BufferSize, PULONGLONG NumberOfBytesRead);
        NtWow64ReadVirtualMemory64.restype = BOOL
    except:
        NtWow64ReadVirtualMemory64 = None

VirtualQueryEx = windll.kernel32.VirtualQueryEx
VirtualQueryEx.argtypes = [HANDLE, LPCVOID, POINTER(MEMORY_BASIC_INFORMATION), c_size_t]
VirtualQueryEx.restype = c_size_t

# VirtualQueryEx64 = windll.kernel32.VirtualQueryEx
# VirtualQueryEx64.argtypes = [HANDLE, LPCVOID, POINTER(MEMORY_BASIC_INFORMATION64), c_size_t]
# VirtualQueryEx64.restype = c_size_t

PAGE_EXECUTE_READWRITE = 64
PAGE_EXECUTE_READ = 32
PAGE_READONLY = 2
PAGE_READWRITE = 4
PAGE_NOCACHE = 512
PAGE_WRITECOMBINE = 1024
PAGE_GUARD = 256

MEM_COMMIT = 4096
MEM_FREE = 65536
MEM_RESERVE = 8192

UNPROTECTED_DACL_SECURITY_INFORMATION = 536870912
DACL_SECURITY_INFORMATION = 4



# Author: Nicolas VERDIER
# This file is part of memorpy.
#
# memorpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# memorpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with memorpy.  If not, see <http://www.gnu.org/licenses/>.

import sys
import string
import re
import logging
import struct
import traceback
import binascii

from .WinProcess import WinProcess as Process
from . import utils
from .Address import Address
from .BaseProcess import ProcessException
from .WinStructures import *

logger = logging.getLogger("memorpy3")

REGEX_TYPE = type(re.compile("^plop$"))


class MemWorker:
    def __init__(
        self, pid=None, name=None, end_offset=None, start_offset=None, debug=True
    ):
        self.process = Process.Process(name=name, pid=pid, debug=debug)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.process.close()

    def Address(self, value, default_type="uint"):
        """ wrapper to instanciate an Address class for the memworker.process"""
        return Address(value, process=self.process, default_type=default_type)

    def umem_replace(self, regex, replace):
        """ like search_replace_mem but works with unicode strings """
        regex = utils.re_to_unicode(regex)
        replace = replace.encode("utf-16-le")
        return self.mem_replace(re.compile(regex, re.UNICODE), replace)

    def mem_replace(self, regex, replace):
        """ search memory for a pattern and replace all found occurrences """
        allWritesSucceed = True
        for _, start_offset in self.mem_search(regex, ftype="re"):
            if self.process.write_bytes(start_offset, replace) == 1:
                logger.debug("Write at offset %s succeeded !" % start_offset)
            else:
                allWritesSucceed = False
                logger.debug("Write at offset %s failed !" % start_offset)

        return allWritesSucceed

    def umem_search(self, regex):
        """ like mem_search but works with unicode strings """
        regex = utils.re_to_unicode(regex)
        for _, i in self.mem_search(str(regex), ftype="re"):
            yield i

    def group_search(self, group, start_offset=None, end_offset=None):
        regex = ""
        for value, type in group:
            if type == "f" or type == "float":
                f = struct.pack("<f", float(value))
                regex += ".." + f[2:4]
            else:
                raise NotImplementedError("unknown type %s" % type)

        return self.mem_search(
            regex, ftype="re", start_offset=start_offset, end_offset=end_offset
        )

    def search_address(self, addr):
        a = "%08X" % addr
        logger.debug("searching address %s" % a)
        regex = ""
        for i in range(len(a) - 2, -1, -2):
            regex += binascii.unhexlify(a[i : i + 2])

        for _, a in self.mem_search(re.escape(regex), ftype="re"):
            yield a

    def parse_re_function(self, b, value, offset):
        for name, regex in value:
            for res in regex.finditer(str(b)):
                yield name, self.Address(offset + res.start(), "bytes")
                """
                index = b.find(res)
                while index != -1:
                    soffset = offset + index
                    if soffset not in duplicates_cache:
                        duplicates_cache.add(soffset)
                        yield name, self.Address(soffset, 'bytes')
                    index = b.find(res, index + len(res))
                """

    def parse_float_function(self, b, value, offset):
        for index in range(0, len(b)):
            try:
                structtype, structlen = utils.type_unpack("float")
                tmpval = struct.unpack(structtype, b[index : index + 4])[0]
                if int(value) == int(tmpval):
                    soffset = offset + index
                    yield self.Address(soffset, "float")
            except Exception as e:
                pass

    def parse_named_groups_function(self, b, value, offset=None):
        for name, regex in value:
            for res in regex.finditer(b):
                yield name, res.groupdict()

    def parse_groups_function(self, b, value, offset=None):
        for name, regex in value:
            for res in regex.finditer(b):
                yield name, res.groups()

    def parse_any_function(self, b, value, offset):
        index = b.find(value)
        while index != -1:
            soffset = offset + index
            yield self.Address(soffset, "bytes")
            index = b.find(value, index + 1)

    def mem_search(
        self,
        value,
        ftype="match",
        protec=PAGE_READWRITE | PAGE_READONLY,
        optimizations=None,
        start_offset=None,
        end_offset=None,
    ):
        """
                iterator returning all indexes where the pattern has been found
        """

        # pre-compile regex to run faster
        if ftype == "re" or ftype == "groups" or ftype == "ngroups":

            # value should be an array of regex
            if type(value) is not list:
                value = [value]

            tmp = []
            for reg in value:
                if type(reg) is tuple:
                    name = reg[0]
                    if type(reg[1]) != REGEX_TYPE:
                        regex = re.compile(reg[1], re.IGNORECASE)
                    else:
                        regex = reg[1]
                elif type(reg) == REGEX_TYPE:
                    name = ""
                    regex = reg
                else:
                    name = ""
                    regex = re.compile(reg, re.IGNORECASE)

                tmp.append((name, regex))
            value = tmp

        elif (
            ftype != "match"
            and ftype != "group"
            and ftype != "re"
            and ftype != "groups"
            and ftype != "ngroups"
            and ftype != "lambda"
        ):
            structtype, structlen = utils.type_unpack(ftype)
            value = struct.pack(structtype, value)

        # different functions avoid if statement before parsing the buffer
        if ftype == "re":
            func = self.parse_re_function

        elif ftype == "groups":
            func = self.parse_groups_function

        elif ftype == "ngroups":
            func = self.parse_named_groups_function

        elif ftype == "float":
            func = self.parse_float_function
        elif ftype == "lambda":  # use a custm function
            func = value
        else:
            func = self.parse_any_function

        if not self.process.isProcessOpen:
            raise ProcessException(
                "Can't read_bytes, process %s is not open" % (self.process.pid)
            )

        for offset, chunk_size in self.process.iter_region(
            start_offset=start_offset,
            end_offset=end_offset,
            protec=protec,
            optimizations=optimizations,
        ):
            b = b""
            current_offset = offset
            chunk_read = 0
            chunk_exc = False
            while chunk_read < chunk_size:
                try:
                    b += self.process.read_bytes(current_offset, chunk_size)
                except IOError as e:
                    print(traceback.format_exc())
                    if e.errno == 13:
                        raise
                    else:
                        logger.warning(e)
                    chunk_exc = True
                    break
                except Exception as e:
                    logger.warning(e)
                    chunk_exc = True
                    break
                finally:
                    current_offset += chunk_size
                    chunk_read += chunk_size

            if chunk_exc:
                continue

            if b:
                if ftype == "lambda":
                    for res in func(b, offset):
                        yield res
                else:
                    for res in func(b, value, offset):
                        yield res



# Author: Nicolas VERDIER
# This file is part of memorpy.
#
# memorpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# memorpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with memorpy.  If not, see <http://www.gnu.org/licenses/>.

from ctypes import windll
import time


def start_winforeground_daemon():
    import threading

    t = threading.Thread(target=window_foreground_loop)
    t.daemon = True
    t.start()


def window_foreground_loop(timeout=20):
    """ set the windows python console to the foreground (for example when you are working with a fullscreen program) """
    hwnd = windll.kernel32.GetConsoleWindow()
    HWND_TOPMOST = -1
    SWP_NOMOVE = 2
    SWP_NOSIZE = 1
    while True:
        windll.user32.SetWindowPos(
            hwnd, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE
        )
        time.sleep(timeout)



# Author: Nicolas VERDIER
# This file is part of memorpy.
#
# memorpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# memorpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with memorpy.  If not, see <http://www.gnu.org/licenses/>.

import copy
import time
import struct

from Address import Address


class Locator:
    """
    take a memoryworker and a type to search then you can feed the locator
    with values and it will reduce the addresses possibilities
    """

    def __init__(self, mw, type="unknown", start=None, end=None):
        self.mw = mw
        self.type = type
        self.last_iteration = {}
        self.last_value = None
        self.start = start
        self.end = end

    def find(self, value, erase_last=True):
        return self.feed(value, erase_last)

    def feed(self, value, erase_last=True):
        self.last_value = value
        new_iter = copy.copy(self.last_iteration)
        if self.type == "unknown":
            all_types = [
                "uint",
                "int",
                "long",
                "ulong",
                "float",
                "double",
                "short",
                "ushort",
            ]
        else:
            all_types = [self.type]
        for type in all_types:
            if type not in new_iter:
                try:
                    new_iter[type] = [
                        Address(x, self.mw.process, type)
                        for x in self.mw.mem_search(
                            value, type, start_offset=self.start, end_offset=self.end
                        )
                    ]
                except struct.error:
                    new_iter[type] = []
            else:
                l = []
                for address in new_iter[type]:
                    try:
                        found = self.mw.process.read(address, type)
                        if int(found) == int(value):
                            l.append(Address(address, self.mw.process, type))
                    except Exception as e:
                        pass

                new_iter[type] = l

        if erase_last:
            del self.last_iteration
            self.last_iteration = new_iter
        return new_iter

    def get_addresses(self):
        return self.last_iteration

    def diff(self, erase_last=False):
        return self.get_modified_addr(erase_last)

    def get_modified_addr(self, erase_last=False):
        last = self.last_iteration
        new = self.feed(self.last_value, erase_last=erase_last)
        ret = {}
        for type, l in last.iteritems():
            typeset = set(new[type])
            for addr in l:
                if addr not in typeset:
                    if type not in ret:
                        ret[type] = []
                    ret[type].append(addr)

        return ret



#!/usr/bin/env python
import sys
import os

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

if sys.version_info < (3,):
    sys.exit("requires python 3")

version = "2.0.0"

if sys.argv[-1] == "publish":
    try:
        import wheel

        print("Wheel version: ", wheel.__version__)
    except ImportError:
        print('Wheel library missing. Please run "pip install wheel"')
        sys.exit()
    os.system("python setup.py sdist upload")
    os.system("python setup.py bdist_wheel upload")
    sys.exit()

if sys.argv[-1] == "tag":
    print("Tagging the version on git:")
    os.system("git tag -a %s -m 'version %s'" % (version, version))
    os.system("git push --tags")
    sys.exit()


setup(
    name="memorpy3",
    version=version,
    description="Python 3 port of the memorypy library using ctypes to search/edit windows programs memory",
    author="Nicolas VERDIER, Reid Castner",
    author_email="contact@n1nj4.eu, rbcastner@gmail.com",
    maintainer_email="rbcastner@gmail.com",
    license="BSD",
    url="https://github.com/rbcastner/memorpy",
    include_package_data=True,
    zip_safe=False,
    packages=["memorpy3"],
    install_requires=[],
    platforms=["Windows"],
    long_description=open("README.md").read(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Intended Audience :: Developers',
    ]
)



# Author: Nicolas VERDIER
# This file is part of memorpy.
#
# memorpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# memorpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with memorpy.  If not, see <http://www.gnu.org/licenses/>.

from ctypes import (
    pointer,
    sizeof,
    windll,
    create_string_buffer,
    c_ulong,
    byref,
    GetLastError,
    c_bool,
    WinError,
)
from .WinStructures import *
import copy
import struct
from . import utils
import platform
from .BaseProcess import BaseProcess, ProcessException

psapi = windll.psapi
kernel32 = windll.kernel32
advapi32 = windll.advapi32

IsWow64Process = None
if hasattr(kernel32, "IsWow64Process"):
    IsWow64Process = kernel32.IsWow64Process
    IsWow64Process.restype = c_bool
    IsWow64Process.argtypes = [c_void_p, POINTER(c_bool)]


class WinProcess(BaseProcess):
    def __init__(self, pid=None, name=None, debug=True):
        """ Create and Open a process object from its pid or from its name """
        super(WinProcess, self).__init__()
        if pid:
            self._open(int(pid), debug=debug)

        elif name:
            self._open_from_name(name, debug=debug)
        else:
            raise ValueError(
                "You need to instanciate process with at least a name or a pid"
            )

        if self.is_64bit():
            si = self.GetNativeSystemInfo()
            self.max_addr = si.lpMaximumApplicationAddress
        else:
            si = self.GetSystemInfo()
            self.max_addr = 2147418111
        self.min_addr = si.lpMinimumApplicationAddress

    def __del__(self):
        self.close()

    def is_64bit(self):
        if not "64" in platform.machine():
            return False
        iswow64 = c_bool(False)
        if IsWow64Process is None:
            return False
        if not IsWow64Process(self.h_process, byref(iswow64)):
            raise WinError()
        return not iswow64.value

    @staticmethod
    def list():
        processes = []
        arr = c_ulong * 256
        lpidProcess = arr()
        cb = sizeof(lpidProcess)
        cbNeeded = c_ulong()
        hModule = c_ulong()
        count = c_ulong()
        modname = create_string_buffer(100)
        PROCESS_QUERY_INFORMATION = 0x0400
        PROCESS_VM_READ = 0x0010

        psapi.EnumProcesses(byref(lpidProcess), cb, byref(cbNeeded))
        nReturned = cbNeeded.value / sizeof(c_ulong())

        pidProcess = [i for i in lpidProcess][:nReturned]
        for pid in pidProcess:
            proc = {"pid": int(pid)}
            hProcess = kernel32.OpenProcess(
                PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, False, pid
            )
            if hProcess:
                psapi.EnumProcessModules(
                    hProcess, byref(hModule), sizeof(hModule), byref(count)
                )
                psapi.GetModuleBaseNameA(
                    hProcess, hModule.value, modname, sizeof(modname)
                )
                proc["name"] = modname.value
                kernel32.CloseHandle(hProcess)
            processes.append(proc)
        return processes

    @staticmethod
    def processes_from_name(processName):
        processes = []
        for process in WinProcess.list():
            if processName == process.get("name", None) or (
                process.get("name", "").lower().endswith(".exe")
                and process.get("name", "")[:-4] == processName
            ):
                processes.append(process)

        if len(processes) > 0:
            return processes

    @staticmethod
    def name_from_process(dwProcessId):
        process_list = WinProcess.list()
        for process in process_list:
            if process.pid == dwProcessId:
                return process.get("name", None)

        return False

    def _open(self, dwProcessId, debug=False):
        if debug:
            ppsidOwner = DWORD()
            ppsidGroup = DWORD()
            ppDacl = DWORD()
            ppSacl = DWORD()
            ppSecurityDescriptor = SECURITY_DESCRIPTOR()

            process = kernel32.OpenProcess(262144, 0, dwProcessId)
            advapi32.GetSecurityInfo(
                kernel32.GetCurrentProcess(),
                6,
                0,
                byref(ppsidOwner),
                byref(ppsidGroup),
                byref(ppDacl),
                byref(ppSacl),
                byref(ppSecurityDescriptor),
            )
            advapi32.SetSecurityInfo(
                process,
                6,
                DACL_SECURITY_INFORMATION | UNPROTECTED_DACL_SECURITY_INFORMATION,
                None,
                None,
                ppSecurityDescriptor.dacl,
                ppSecurityDescriptor.group,
            )
            kernel32.CloseHandle(process)
        self.h_process = kernel32.OpenProcess(2035711, 0, dwProcessId)
        if self.h_process is not None:
            self.isProcessOpen = True
            self.pid = dwProcessId
            return True
        return False

    def close(self):
        if self.h_process is not None:
            ret = kernel32.CloseHandle(self.h_process) == 1
            if ret:
                self.h_process = None
                self.pid = None
                self.isProcessOpen = False
            return ret
        return False

    def _open_from_name(self, processName, debug=False):
        processes = self.processes_from_name(processName)
        if not processes:
            raise ProcessException("can't get pid from name %s" % processName)
        elif len(processes) > 1:
            raise ValueError(
                "There is multiple processes with name %s. Please select a process from its pid instead"
                % processName
            )
        if debug:
            self._open(processes[0]["pid"], debug=True)
        else:
            self._open(processes[0]["pid"], debug=False)

    def GetSystemInfo(self):
        si = SYSTEM_INFO()
        kernel32.GetSystemInfo(byref(si))
        return si

    def GetNativeSystemInfo(self):
        si = SYSTEM_INFO()
        kernel32.GetNativeSystemInfo(byref(si))
        return si

    def VirtualQueryEx(self, lpAddress):
        mbi = MEMORY_BASIC_INFORMATION()
        if not VirtualQueryEx(self.h_process, lpAddress, byref(mbi), sizeof(mbi)):
            raise ProcessException("Error VirtualQueryEx: 0x%08X" % lpAddress)
        return mbi

    def VirtualQueryEx64(self, lpAddress):
        mbi = MEMORY_BASIC_INFORMATION64()
        if not VirtualQueryEx64(self.h_process, lpAddress, byref(mbi), sizeof(mbi)):
            raise ProcessException("Error VirtualQueryEx: 0x%08X" % lpAddress)
        return mbi

    def VirtualProtectEx(self, base_address, size, protection):
        old_protect = c_ulong(0)
        if not kernel32.VirtualProtectEx(
            self.h_process, base_address, size, protection, byref(old_protect)
        ):
            raise ProcessException(
                "Error: VirtualProtectEx(%08X, %d, %08X)"
                % (base_address, size, protection)
            )
        return old_protect.value

    def iter_region(
        self, start_offset=None, end_offset=None, protec=None, optimizations=None
    ):

        offset = start_offset or self.min_addr
        end_offset = end_offset or self.max_addr

        while True:
            if offset >= end_offset:
                break
            mbi = self.VirtualQueryEx(offset)
            offset = mbi.BaseAddress
            chunk = mbi.RegionSize
            protect = mbi.Protect
            state = mbi.State
            # print "offset: %s, chunk:%s"%(offset, chunk)
            if state & MEM_FREE or state & MEM_RESERVE:
                offset += chunk
                continue
            if protec:
                if (
                    not protect & protec
                    or protect & PAGE_NOCACHE
                    or protect & PAGE_WRITECOMBINE
                    or protect & PAGE_GUARD
                ):
                    offset += chunk
                    continue
            yield offset, chunk
            offset += chunk

    def write_bytes(self, address, data):
        address = int(address)
        if not self.isProcessOpen:
            raise ProcessException(
                "Can't write_bytes(%s, %s), process %s is not open"
                % (address, data, self.pid)
            )
        buffer = create_string_buffer(data)
        sizeWriten = c_size_t(0)
        bufferSize = sizeof(buffer) - 1
        _address = address
        _length = bufferSize + 1
        try:
            old_protect = self.VirtualProtectEx(
                _address, _length, PAGE_EXECUTE_READWRITE
            )
        except:
            pass

        res = kernel32.WriteProcessMemory(
            self.h_process, address, buffer, bufferSize, byref(sizeWriten)
        )
        try:
            self.VirtualProtectEx(_address, _length, old_protect)
        except:
            pass

        return res

    def read_bytes(self, address, bytes=4, use_NtWow64ReadVirtualMemory64=False):
        # print "reading %s bytes from addr %s"%(bytes, address)
        if use_NtWow64ReadVirtualMemory64:
            if NtWow64ReadVirtualMemory64 is None:
                raise WindowsError(
                    "NtWow64ReadVirtualMemory64 is not available from a 64bit process"
                )
            RpM = NtWow64ReadVirtualMemory64
        else:
            RpM = ReadProcessMemory

        address = int(address)
        buffer = create_string_buffer(bytes)
        bytesread = c_size_t(0)
        data = b""
        length = bytes
        while length:
            if RpM(self.h_process, address, buffer, bytes, byref(bytesread)) or (
                use_NtWow64ReadVirtualMemory64 and GetLastError() == 0
            ):
                if bytesread.value:
                    data += buffer.raw[: bytesread.value]
                    length -= bytesread.value
                    address += bytesread.value
                if not len(data):
                    raise ProcessException(
                        "Error %s in ReadProcessMemory(%08x, %d, read=%d)"
                        % (GetLastError(), address, length, bytesread.value)
                    )
                return data
            else:
                if (
                    GetLastError() == 299
                ):  # only part of ReadProcessMemory has been done, let's return it
                    data += buffer.raw[: bytesread.value]
                    return data
                raise WinError()
            # data += buffer.raw[:bytesread.value]
            # length -= bytesread.value
            # address += bytesread.value
        return data

    def list_modules(self):
        module_list = []
        if self.pid is not None:
            hModuleSnap = CreateToolhelp32Snapshot(TH32CS_CLASS.SNAPMODULE, self.pid)
            if hModuleSnap is not None:
                module_entry = MODULEENTRY32()
                module_entry.dwSize = sizeof(module_entry)
                success = Module32First(hModuleSnap, byref(module_entry))
                while success:
                    if module_entry.th32ProcessID == self.pid:
                        module_list.append(copy.copy(module_entry))
                    success = Module32Next(hModuleSnap, byref(module_entry))

                kernel32.CloseHandle(hModuleSnap)
        return module_list

    def get_symbolic_name(self, address):
        for m in self.list_modules():
            if int(m.modBaseAddr) <= int(address) < int(m.modBaseAddr + m.modBaseSize):
                return "%s+0x%08X" % (m.szModule, int(address) - m.modBaseAddr)

        return "0x%08X" % int(address)

    def hasModule(self, module):
        if module[-4:] != ".dll":
            module += ".dll"
        module_list = self.list_modules()
        for m in module_list:
            if module in m.szExePath.split("\\"):
                return True
        return False

    def get_instruction(self, address):
        """
        Pydasm disassemble utility function wrapper. Returns the pydasm decoded instruction in self.instruction.
        """
        import pydasm

        try:
            data = self.read_bytes(int(address), 32)
        except:
            return "Unable to disassemble at %08x" % address

        return pydasm.get_instruction(data, pydasm.MODE_32)
